from functools import partial
from collections import defaultdict
from time import time, sleep
from itertools import count as itercount
try:
    from itertools import imap
except ImportError:
    # In python3, map is now lazy.
    imap = map

try:
  #Python2
  rawtype = str
  safetype = unicode
  raw2str = lambda r: r.decode('utf-8')
  str2raw = lambda s: s.encode('utf-8')
except:
  #Python3
  rawtype = bytes
  safetype = str
  raw2str = lambda r: r.decode('utf-8')
  str2raw = lambda s: s.encode('utf-8')

def ingest(d):
  """Convert rawtype (str py27 or bytes py3x) to safetype (unicode py27 or str py3x)"""
  if isinstance(d, rawtype):
    return raw2str(d)
  elif isinstance(d, safetype):
    return d
  else:
    raise TypeError("Can't ingest data of type %s" % type(d))

def egest(s):
  """Convert safetype (unicode py27, str py3x) to rawtype (str py27 or bytes py3x)"""
  if isinstance(s, rawtype):
    return s
  elif isinstance(s, safetype): # On python 2 str is bytes.
    return str2raw(s)
  else:
    raise TypeError("Can't egest data of type %s" % type(s))

""""
If zero length array is passed, returns ["."].
Otherwise returns the origional array.
Useful for dealing with lists of files, or not.
"""
def empty2dot(paths):
  if len(paths) == 0:
    return ["."]
  else:
    return paths

def compose(f, g):
  return lambda *args, **kwargs: f(g(*args, **kwargs))

def composeFunctor(f,g):
    return lambda x: f(g(x))

def concat(l):
  for sublist in l:
    for item in sublist:
      yield item

def concatMap(func):
  return compose(concat, partial(imap, func))

def fmap(func):
  def mapped(collection):
    return imap(func, collection)
  return mapped

def identity(x):
    return x

def groupby(func, l):
  groups = defaultdict(list)
  for i in l:
    groups[func(i)].append(i)
  return list(groups.items())

def take(count):
  def taker(collection):
    remaining = count
    i = iter(collection)
    while remaining > 0:
      try:
        yield next(i)
      except StopIteration:
        return
      remaining = remaining - 1
  return taker

def consume(collection):
  for _ in collection:
    pass

def uniq(l):
  seen = set()
  for i in l:
    if i in seen:
      continue
    else:
      seen.add(i)
      yield i

def irange(start, increment):
  while True :
    yield start
    start += increment

def invert(v):
    return not(v)

#TODO why not len?
def count(iterator):
    c = 0
    for v in iterator:
        c+=1
    return c

def uncurry(func):
  """Wraps func so that the first arg is expanded into list args."""
  def uncurried(list_args, **kwargs):
    return func(*list_args, **kwargs)
  return uncurried

def curry(func):
  """"Wraps func so that a series of args are turned into a single arg list."""
  def curried(*args, **kwargs):
    return func(args, **kwargs)
  return curried

def identify(func):
  """Wrap func so that it returns what comes in."""
  def identified(arg):
    func(arg)
    return arg
  return identified

def pipeline(*funcs):
  if funcs:
    foo = funcs[0]
    rest = funcs[1:]
    if rest:
      next_hop = pipeline(*rest)
      def pipe(*args, **kwargs):
        return next_hop(foo(*args, **kwargs))
      return pipe
    else: # no rest, foo is final function.
      return foo
  else: # no funcs at all.
    return fmap(identity)

def zipFrom(a, bs):
    """Converts a value and list into a list of tuples: a -> [b] -> [(a,b)]"""
    for b in bs:
        yield (a, b)

def dot(fn):
    """Reverses the dot syntax (object.attr), so you can do dot(attr)(obj)."""
    def access(obj):
        return getattr(obj, fn)
    return access

def nth(n):
    def nth_getter(lst):
        return lst[n]
    return nth_getter

first = nth(0)
second = nth(1)

def maybe(default, v):
    if v:
        return v
    else:
        return default

def every(predicate, coll):
    for x in coll:
        if not predicate(x):
            return False
    return True

def repeater(callback, period=0, max_tries=None, max_time=None, predicate = identity, catch_predicate = lambda e: False):
  def repeat_worker(*args, **kwargs):
    if max_time is not None:
      deadline = time() + max_time
    else:
      deadline = None
    if max_tries is None:
        r = itercount()
    else:
        r = range(0, max_tries)
    for i in r:
      start_time = time()
      try:
        ret = callback(*args, **kwargs)
      except Exception as e:
        if not catch_predicate(e):
          return False
      else:
        if predicate(ret):
          return True
        if deadline is not None and time() > deadline:
          return False
        end_time = time()
        sleep_time = max(0.0, period - (end_time - start_time))
        sleep(sleep_time)
    return False
  return repeat_worker



from __future__ import print_function
import farmfs
from farmfs import getvol
from docopt import docopt
from functools import partial
from farmfs import cwd
from farmfs.util import empty2dot, fmap, pipeline, concat, identify, uncurry, count, groupby, consume, concatMap, zipFrom, safetype, ingest, first, maybe, every, identity, repeater, uniq
from farmfs.volume import mkfs, tree_diff, tree_patcher, encode_snapshot, blob_import
from farmfs.fs import Path, userPath2Path, ftype_selector, FILE, LINK, skip_ignored, is_readonly, ensure_symlink
from json import JSONEncoder
import sys
try:
    from itertools import ifilter
except ImportError:
    # On python3, filter is lazy.
    ifilter = filter
try:
    from itertools import imap
except ImportError:
    # On python3 map is lazy.
    imap = map

json_encode = lambda data: JSONEncoder(ensure_ascii=False, sort_keys=True).encode(data)
json_printr = pipeline(list, json_encode, print)
strs_printr = pipeline(fmap(print), consume)

def dict_printr(keys, d):
    print("\t".join([ingest(d.get(k, '')) for k in keys]))

def dicts_printr(keys):
    return pipeline(fmap(partial(dict_printr, keys)), consume)

snapshot_printr = dicts_printr(['path', 'type', 'csum'])

UI_USAGE = \
"""
FarmFS

Usage:
  farmfs mkfs [--root <root>] [--data <data>]
  farmfs (status|freeze|thaw) [<path>...]
  farmfs snap list
  farmfs snap (make|read|delete|restore|diff) <snap>
  farmfs fsck [--broken --frozen-ignored --blob-permissions --checksums]
  farmfs count
  farmfs similarity
  farmfs gc [--noop]
  farmfs remote add <remote> <root>
  farmfs remote remove <remote>
  farmfs remote list [<remote>]
  farmfs pull <remote> [<snap>]
  farmfs diff <remote> [<snap>]


Options:

"""

def op_doer(op):
    (blob_op, tree_op, desc) = op
    blob_op()
    tree_op()

stream_op_doer = fmap(op_doer)


def fsck_missing_blobs(vol, cwd):
    '''Look for blobs in tree or snaps which are not in blobstore.'''
    trees = vol.trees()
    tree_items = concatMap(lambda t: zipFrom(t,iter(t)))
    tree_links = partial(ifilter, uncurry(lambda snap, item: item.is_link()))
    broken_tree_links = partial(
            ifilter,
            uncurry(lambda snap, item: not vol.blob_checker(item.csum())))
    checksum_grouper = partial(groupby,
            uncurry(lambda snap, item: item.csum()))
    def broken_link_printr(csum, snap_items):
        print(csum)
        for (snap, item) in snap_items:
            print(
                    "\t",
                    snap.name,
                    item.to_path(vol.root).relative_to(cwd, leading_sep=False))
    broken_links_printr = fmap(identify(uncurry(broken_link_printr)))
    num_bad_blobs = pipeline(
            tree_items,
            tree_links,
            broken_tree_links,
            checksum_grouper,
            broken_links_printr,
            count)(trees)
    return num_bad_blobs

def fsck_frozen_ignored(vol, cwd):
    '''Look for frozen links which are in the ignored file.'''
    ignore_mdd = partial(skip_ignored, [safetype(vol.mdd)])
    ignored_frozen = pipeline(
            ftype_selector([LINK]),
            partial(ifilter, uncurry(vol.is_ignored)),
            fmap(first),
            fmap(lambda p: p.relative_to(cwd, leading_sep=False)),
            fmap(partial(print, "Ignored file frozen")),
            count
            )(vol.root.entries(ignore_mdd))
    return ignored_frozen

def fsck_blob_permissions(vol, cwd):
    '''Look for blobstore blobs which are not readonly.'''
    blob_permissions = pipeline(
            partial(ifilter, is_readonly),
            fmap(vol.reverser),
            fmap(partial(print, "writable blob: ")),
            count
            )(vol.userdata_files())
    return blob_permissions

def fsck_checksum_mismatches(vol, cwd):
    '''Look for checksum mismatches.'''
    select_broken = partial(ifilter, vol.check_userdata_blob)
    #TODO CORRUPTION checksum mismatch in blob <CSUM>, would be nice to know back references.
    mismatches = pipeline(
            select_broken,
            fmap(vol.reverser),
            fmap(lambda csum: print("CORRUPTION checksum mismatch in blob %s" % csum)),
            count
            )(vol.userdata_files())
    return mismatches

def ui_main():
    result = farmfs_ui(sys.argv[1:], cwd)
    exit(result)

def farmfs_ui(argv, cwd):
  exitcode = 0
  args = docopt(UI_USAGE, argv)
  if args['mkfs']:
    root = userPath2Path(args['<root>'] or ".", cwd)
    data = userPath2Path(args['<data>'], cwd) if args.get('<data>') else Path(".farmfs/userdata", root)
    mkfs(root, data)
    print("FileSystem Created %s using blobstore %s" % (root, data))
  else:
    vol = getvol(cwd)
    paths = map(lambda x: userPath2Path(x, cwd), empty2dot(args['<path>']))
    def delta_printr(delta):
      deltaPath = delta.path(vol.root).relative_to(cwd, leading_sep=False)
      print("diff: %s %s %s" % (delta.mode, deltaPath, delta.csum))
    stream_delta_printr = fmap(identify(delta_printr))
    def op_printr(op):
      (blob_op, tree_op, (desc, path)) = op
      print(desc % path.relative_to(cwd, leading_sep=False))
    stream_op_printr = fmap(identify(op_printr))
    if args['status']:
      get_thawed = fmap(vol.thawed)
      pipeline(get_thawed,
              concat,
              fmap(lambda p: p.relative_to(cwd, leading_sep=False)),
              fmap(print),
              consume)(paths)
    elif args['freeze']:
      def printr(freeze_op):
        s = "Imported %s with checksum %s" % \
                (freeze_op['path'].relative_to(cwd, leading_sep=False),
                 freeze_op['csum'])
        if freeze_op['was_dup']:
          print(s, "was a duplicate")
        else:
          print(s)
      importer = fmap(vol.freeze)
      get_thawed = fmap(vol.thawed)
      print_list = fmap(printr)
      pipeline(get_thawed, concat, importer, print_list, consume)(paths)
    elif args['thaw']:
      def printr(path):
        print("Exported %s" % path.relative_to(cwd, leading_sep=False))
      exporter = fmap(vol.thaw)
      get_frozen = fmap(vol.frozen)
      print_list = fmap(printr)
      pipeline(get_frozen, concat, exporter, print_list, consume)(paths)
    elif args['fsck']:
        fsck_actions = {
                '--broken': (fsck_missing_blobs, 1),
                '--frozen-ignored': (fsck_frozen_ignored, 4),
                '--blob-permissions': (fsck_blob_permissions, 8),
                '--checksums': (fsck_checksum_mismatches, 2),
                }
        fsck_tasks = [action for (verb, action) in fsck_actions.items() if args[verb]]
        if len(fsck_tasks) == 0:
            # No options were specified, run the whole sweet.
            fsck_tasks = fsck_actions.values()
        for foo, fail_code in fsck_tasks:
            exitcode = exitcode | (foo(vol, cwd) and fail_code)
    elif args['count']:
      trees = vol.trees()
      tree_items = concatMap(lambda t: zipFrom(t,iter(t)))
      tree_links = partial(ifilter, uncurry(lambda snap, item: item.is_link()))
      checksum_grouper = partial(groupby,
              uncurry(lambda snap, item: item.csum()))
      def count_printr(csum, snap_items):
        print(csum, count(snap_items))
        for (snap, item) in snap_items:
            print(snap.name, item.to_path(vol.root).relative_to(cwd, leading_sep=False))
      counts_printr = fmap(identify(uncurry(count_printr)))
      pipeline(
              tree_items,
              tree_links,
              checksum_grouper,
              counts_printr,
              consume
              )(trees)
    elif args['similarity']:
      for (dir_a, count_a, dir_b, count_b, intersect) in vol.similarity():
        assert isinstance(dir_a, Path)
        assert isinstance(dir_b, Path)
        path_a = dir_a.relative_to(cwd, leading_sep=False)
        path_b = dir_b.relative_to(cwd, leading_sep=False)
        print(path_a, "%d/%d %d%%" % (intersect, count_a, int(100*float(intersect)/count_a)), \
                path_b, "%d/%d %d%%" % (intersect, count_b, int(100*float(intersect)/count_b)))
    elif args['gc']:
      applyfn = fmap(identity) if args.get('--noop') else fmap(vol.delete_blob)
      fns = [fmap(identify(partial(print, "Removing"))),
              applyfn,
              consume]
      pipeline(*fns)(sorted(vol.unused_blobs(vol.items())))
    elif args['snap']:
      snapdb = vol.snapdb
      if args['list']:
        #TODO have an optional argument for which remote.
        print("\n".join(snapdb.list()))
      else:
        name = args['<snap>']
        if args['delete']:
          snapdb.delete(name)
        elif args['make']:
          snapdb.write(name, vol.tree())
        else:
          snap = snapdb.read(name)
          if args['read']:
            for i in snap:
              print(i)
          elif args['restore']:
            tree = vol.tree()
            diff = tree_diff(vol.tree(), snap)
            pipeline(
                    stream_delta_printr,
                    tree_patcher(vol, vol),
                    stream_op_printr,
                    stream_op_doer,
                    consume)(diff)
          elif args['diff']:
            diff = tree_diff(vol.tree(), snap)
            pipeline(stream_delta_printr, consume)(diff)
    elif args['remote']:
      if args["add"]:
        remote_vol = getvol(userPath2Path(args['<root>'], cwd))
        vol.remotedb.write(args['<remote>'], remote_vol)
      elif args["remove"]:
        vol.remotedb.delete(args['<remote>'])
      elif args["list"]:
        if args["<remote>"]:
          remote_vol = vol.remotedb.read(args['<remote>'])
          print("\n".join(remote_vol.snapdb.list()))
        else:
          for remote_name in vol.remotedb.list():
            remote_vol = vol.remotedb.read(remote_name)
            print(remote_name, remote_vol.root)
    elif args['pull'] or args['diff']:
      remote_vol = vol.remotedb.read(args['<remote>'])
      snap_name = args['<snap>']
      remote_snap = remote_vol.snapdb.read(snap_name) if snap_name else remote_vol.tree()
      diff = tree_diff(vol.tree(), remote_snap)
      if args['pull']:
        patcher = tree_patcher(vol, remote_vol)
        pipeline(
                stream_delta_printr,
                patcher,
                stream_op_printr,
                stream_op_doer,
                consume)(diff)
      else: # diff
        pipeline(stream_delta_printr, consume)(diff)
  return exitcode


def printNotNone(value):
  if value is not None:
    print(value)

def walk(parents, is_ignored, match):
  return pipeline(
          concatMap(lambda parent: parent.entries(is_ignored)),
          ftype_selector(match)
          )(iter(parents))

def reverse(vol, csum):
  """Yields a set of paths which reference a given checksum_path name."""

DBG_USAGE = \
"""
FarmDBG

Usage:
  farmdbg reverse <csum>
  farmdbg key read <key>
  farmdbg key write <key> <value>
  farmdbg key delete <key>
  farmdbg key list [<key>]
  farmdbg walk (keys|userdata|root|snap <snapshot>) [--json]
  farmdbg checksum <path>...
  farmdbg fix link [--remote=<remote>] <target> <file>
  farmdbg rewrite-links <target>
  farmdbg missing <snap>...
  farmdbg blobtype <blob>...
  farmdbg blob <blob>...
  farmdbg s3 list <bucket> <prefix>
  farmdbg s3 upload <bucket> <prefix>
"""

from s3lib import Connection as s3conn
from s3lib.ui import load_creds as load_s3_creds
def dbg_main():
  return dbg_ui(sys.argv[1:], cwd)

def dbg_ui(argv, cwd):
  exitcode = 0
  args = docopt(DBG_USAGE, argv)
  vol = getvol(cwd)
  if args['reverse']:
    csum = args['<csum>']
    trees = vol.trees()
    tree_items = concatMap(lambda t: zipFrom(t,iter(t)))
    tree_links = partial(ifilter, uncurry(lambda snap, item: item.is_link()))
    matching_links = partial(ifilter, uncurry(lambda snap, item: item.csum() == csum))
    def link_printr(snap_item):
        (snap, item) = snap_item
        print(snap.name, item.to_path(vol.root).relative_to(cwd, leading_sep=False))
    links_printr = fmap(identify(link_printr))
    pipeline(
            tree_items,
            tree_links,
            matching_links,
            links_printr,
            consume)(trees)
  elif args['key']:
    db = vol.keydb
    key = args['<key>']
    if args['read']:
      printNotNone(db.readraw(key))
    elif args['delete']:
      db.delete(key)
    elif args['list']:
      for v in db.list(key):
        print(v)
    elif args['write']:
      value = args['<value>']
      db.write(key, value)
  elif args['walk']:
    if args['root']:
      printr = json_printr if args.get('--json') else snapshot_printr
      printr(encode_snapshot(vol.tree()))
    elif args['snap']:
      printr = json_printr if args.get('--json') else snapshot_printr
      printr(encode_snapshot(vol.snapdb.read(args['<snapshot>'])))
    elif args['userdata']:
      printr = json_printr if args.get('--json') else strs_printr
      userdata = pipeline(
              fmap(first),
              fmap(vol.reverser),
              ) (walk([vol.udd], None, [FILE]))
      printr(userdata)
    elif args['keys']:
      printr = json_printr if args.get('--json') else strs_printr
      printr(vol.keydb.list())
  elif args['checksum']:
    #TODO <checksum> <full path>
    paths = imap(lambda x: Path(x, cwd), empty2dot(args['<path>']))
    for p in paths:
      print(p.checksum(), p.relative_to(cwd, leading_sep=False))
  elif args['link']:
    #TODO might move into blobstore.
    f = Path(args['<file>'], cwd)
    b = ingest(args['<target>'])
    bp = vol.csum_to_path(b)
    if not bp.exists():
      print("blob %s doesn't exist" % b)
      if args['--remote']:
        remote = vol.remotedb.read(args['--remote'])
      else:
        raise(ValueError("aborting due to missing blob"))
      rbp = remote.csum_to_path(b)
      blob_import(rbp, bp)
    else:
      pass #bp exists, can we check its checksum?
    ensure_symlink(f, bp)
  elif args['rewrite-links']:
    target = Path(args['<target>'], cwd)
    for (link, _type) in walk([target], [safetype(vol.mdd)], [LINK]):
      new = vol.repair_link(link)
      if new is not None:
          print("Relinked %s to %s" % (link.relative_to(cwd, leading_sep=False), new))
  elif args['missing']:
    tree_csums = pipeline(
            partial(ifilter, lambda item: item.is_link()),
            fmap(lambda item: item.csum()),
            set
            )(iter(vol.tree()))
    snapNames = args['<snap>']
    def missing_printr(csum, pathStrs):
        paths = sorted(imap(lambda pathStr: vol.root.join(pathStr), pathStrs))
        for path in paths:
            print("%s\t%s" % (csum, path.relative_to(cwd, leading_sep=False)))
    missing_csum2pathStr = pipeline(
            fmap(vol.snapdb.read),
            concatMap(iter),
            partial(ifilter, lambda item: item.is_link()),
            partial(ifilter, lambda item: not vol.is_ignored(item.to_path(vol.root), None)),
            partial(ifilter, lambda item: item.csum() not in tree_csums),
            partial(groupby, lambda item: item.csum()),
            partial(ifilter, uncurry(lambda csum, items: every(lambda item: not item.to_path(vol.root).exists(), items))),
            fmap(uncurry(lambda csum, items: (csum, list(imap(lambda item: item.pathStr(), items))))),
            fmap(uncurry(missing_printr)),
            count
            )(snapNames)
  elif args['blobtype']:
    for blob in args['<blob>']:
      blob = ingest(blob)
      print(
              blob,
              maybe("unknown", vol.csum_to_path(blob).filetype()))
  elif args['blob']:
    for csum in args['<blob>']:
      csum = ingest(csum)
      print(csum,
              vol.csum_to_path(csum).relative_to(cwd, leading_sep=False))
  elif args['s3']:
      bucket = args['<bucket>']
      prefix = args['<prefix>'] + "/"
      access_id, secret_key = load_s3_creds(None)
      with s3conn(access_id, secret_key) as s3:
          key_iter = s3.list_bucket(bucket, prefix=prefix)
          if args['list']:
              pipeline(fmap(print), consume)(key_iter)
          elif args['upload']:
              keys = set(key_iter)
              tree = vol.tree()
              def upload(csum):
                  blob = vol.csum_to_path(csum)
                  key = prefix + csum
                  print(csum, "->", blob, "->", key)
                  with blob.open('rb') as f:
                      #TODO should provide pre-calculated md5 rather than recompute.
                      result = s3.put_object(bucket, key, f.read())
                  return result
              http_success = lambda status_headers: status_headers[0] >=200 and status_headers[0] < 300
              s3_exception = lambda e: isinstance(e, ValueError)
              upload_repeater = repeater(upload, max_tries = 3, predicate = http_success, catch_predicate = s3_exception)
              pipeline(
                      partial(ifilter, lambda x: x.is_link()),
                      fmap(lambda x: x.csum()),
                      partial(ifilter, lambda x: x not in keys),
                      uniq,
                      fmap(upload_repeater),
                      consume
                      )(iter(tree))
  return exitcode



from farmfs.fs import Path
from farmfs.fs import ensure_dir
from farmfs.fs import ensure_file
from farmfs.fs import Path
from hashlib import md5
from json import loads, JSONEncoder
from errno import ENOENT as NoSuchFile
from errno import EISDIR as IsDirectory
from os.path import sep
from func_prototypes import typed, returned
from farmfs.util import ingest, egest, safetype

@returned(str)
@typed(bytes)
def checksum(value_bytes):
  """Input string should already be coersed into an encoding before being provided"""
  return md5(value_bytes).hexdigest()

class KeyDB:
  def __init__(self, db_path):
    assert isinstance(db_path, Path)
    self.root = db_path

  #TODO I DONT THINK THIS SHOULD BE A PROPERTY OF THE DB UNLESS WE HAVE SOME ITERATOR BASED RECORD TYPE.
  def write(self, key, value):
    key = safetype(key)
    value_json = JSONEncoder(ensure_ascii=False, sort_keys=True).encode(value)
    value_bytes = egest(value_json)
    value_hash = egest(checksum(value_bytes))
    key_path = self.root.join(key)
    with ensure_file(key_path, 'wb') as f:
      f.write(value_bytes)
      f.write(b"\n")
      f.write(value_hash)
      f.write(b"\n")

  def readraw(self, key):
    key = safetype(key)
    try:
      with self.root.join(key).open('rb') as f:
        obj_bytes = f.readline().strip()
        obj_bytes_checksum = checksum(obj_bytes).encode('utf-8')
        key_checksum = f.readline().strip()
      if obj_bytes_checksum != key_checksum:
        raise ValueError("Checksum mismatch for key %s. Expected %s, calculated %s" % (key, key_checksum, obj_bytes_checksum))
      obj_str = egest(obj_bytes)
      return obj_str
    except IOError as e:
      if e.errno == NoSuchFile or e.errno == IsDirectory:
        return None
      else:
        raise e

  def read(self, key):
    obj_str = self.readraw(key)
    if obj_str is None:
      return None
    else:
      obj = loads(obj_str)
      return obj

  def list(self, query=None):
    if query is None:
      query = ""
    query = safetype(query)
    query_path = self.root.join(query)
    assert self.root in query_path.parents(), "%s is not a parent of %s" % (self.root, query_path)
    if query_path.exists and query_path.isdir():
      return [ p.relative_to(self.root, leading_sep=False)
          for (p,t) in query_path.entries()
          if t == 'file' ]
    else:
      return []

  def delete(self, key):
    key = safetype(key)
    path = self.root.join(key)
    path.unlink(clean=self.root)

class KeyDBWindow(KeyDB):
  def __init__(self, window, keydb):
    window = safetype(window)
    assert isinstance(keydb, KeyDB)
    self.prefix = window + sep
    self.keydb = keydb

  def write(self, key, value):
    assert(key)
    assert(value)
    self.keydb.write(self.prefix+key, value)

  def read(self, key):
    return self.keydb.read(self.prefix+key)

  def list(self,):
    return [ x[len(self.prefix):] for x in self.keydb.list(self.prefix) ]

  def delete(self, key):
    self.keydb.delete(self.prefix+key)

class KeyDBFactory():
  def __init__(self, keydb, encoder, decoder):
    self.keydb = keydb
    self.encoder = encoder
    self.decoder = decoder

  def write(self, key, value):
    self.keydb.write(key, self.encoder(value))

  def read(self, key):
    return self.decoder(self.keydb.read(key), key)

  def list(self,):
    return self.keydb.list()

  def delete(self, key):
    self.keydb.delete(key)

  #TODO I don't think I used this.
  def copy(self, key, remote):
    value = remote.read(key)
    self.write(key, value)



from farmfs.volume import mkfs as make_volume
from farmfs.volume import FarmFSVolume
from farmfs.fs import Path
from farmfs.keydb import KeyDBWindow
from func_prototypes import typed, returned
from farmfs.util import take, ingest
try:
    from os import getcwdu
    getcwd_utf = lambda : ingest(getcwdu())
except ImportError:
    from os import getcwdb
    getcwd_utf = lambda : ingest(getcwdb())
try:
    from itertools import imap
except ImportError:
    # On python3 map is lazy.
    imap = map
try:
    from itertools import ifilter
except ImportError:
    ifilter = filter

cwd = Path(getcwd_utf())

@returned(Path)
@typed(Path)
def _find_root_path(path):
  candidates = imap(lambda x: x.join(".farmfs"), path.parents())
  matches = ifilter(lambda x: x.isdir(), candidates)
  root = next(take(1)(matches), None)
  if root:
    nested_root = next(take(1)(matches), None)
    if nested_root:
      raise ValueError("Farmfs volumes cannot be nested")
    return root.parent()
  else:
   raise ValueError("Volume not found: %s" % path)

@returned(FarmFSVolume)
@typed(Path)
def getvol(path):
  root = _find_root_path(path)
  vol = FarmFSVolume(root)
  return vol




from errno import ENOENT as NoSuchFile
from farmfs.keydb import KeyDB
from farmfs.keydb import KeyDBWindow
from farmfs.keydb import KeyDBFactory
from farmfs.util import *
from farmfs.fs import Path
from farmfs.fs import ensure_absent, ensure_link, ensure_symlink, ensure_readonly, ensure_copy, ensure_dir, skip_ignored, ftype_selector, FILE, LINK, DIR
from farmfs.snapshot import Snapshot, TreeSnapshot, KeySnapshot, SnapDelta, encode_snapshot, decode_snapshot
from os.path import sep
from itertools import combinations, chain
try:
    from itertools import imap
except ImportError:
    # On python3 map is lazy.
    imap = map
from func_prototypes import typed, returned
from functools import partial
try:
    from itertools import ifilter
except ImportError:
    ifilter = filter
import re

def _metadata_path(root):
  assert isinstance(root, Path)
  return root.join(".farmfs")

@returned(Path)
@typed(Path)
def _keys_path(root):
  return _metadata_path(root).join("keys")

@returned(Path)
@typed(Path)
def _snaps_path(root):
  return _metadata_path(root).join("snaps")

def mkfs(root, udd):
  assert isinstance(root, Path)
  assert isinstance(udd, Path)
  root.mkdir()
  _metadata_path(root).mkdir()
  _keys_path(root).mkdir()
  _snaps_path(root).mkdir()
  kdb = KeyDB(_keys_path(root))
  # Make sure root key is removed.
  kdb.delete("root")
  kdb.write('udd', safetype(udd))
  udd.mkdir()
  kdb.write('status', {})
  vol = FarmFSVolume(root)

@returned(safetype)
@typed(safetype, int, int)
def _checksum_to_path(checksum, num_segs=3, seg_len=3):
  segs = [ checksum[i:i+seg_len] for i in range(0, min(len(checksum), seg_len * num_segs), seg_len)]
  segs.append(checksum[num_segs*seg_len:])
  return sep.join(segs)

_sep_replace_ = re.compile(sep)
@returned(safetype)
@typed(safetype)
def _remove_sep_(path):
    return _sep_replace_.subn("",path)[0]

def reverser(num_segs=3):
  """Returns a function which takes Paths into the user data and returns csums."""
  r = re.compile("((\/([0-9]|[a-f])+){%d})$" % (num_segs+1))
  def checksum_from_link(link):
    """Takes a path into the userdata, returns the matching csum."""
    m = r.search(safetype(link))
    if (m):
      csum_slash = m.group()[1:]
      csum = _remove_sep_(csum_slash)
      return csum
    else:
      raise ValueError("link %s checksum didn't parse" %(link))
  return checksum_from_link

def _validate_checksum(link2csum, path):
  csum = path.checksum()
  link_csum = link2csum(path)
  return csum == link_csum

def directory_signatures(snap, root):
  dirs = {}
  for entry in snap:
    if entry.is_link():
      (path_str, _, csum) = entry.get_tuple()
      parent = root.join(path_str).parent()
      try:
        dirs[parent].update([csum])
      except KeyError:
        dirs[parent] = set([csum])
  return dirs

def encode_volume(vol):
  return safetype(vol.root)

def decode_volume(vol, key):
  return FarmFSVolume(Path(vol))

def encode_snapshot(snap):
  return list(imap(lambda x: x.get_dict(), snap))

def decode_snapshot(reverser, data, key):
  return KeySnapshot(data, key, reverser)

class FarmFSVolume:
  def __init__(self, root):
    assert isinstance(root, Path)
    self.root = root
    self.mdd = _metadata_path(root)
    self.keydb = KeyDB(_keys_path(root))
    self.udd = Path(self.keydb.read('udd'))
    self.reverser = reverser()
    self.snapdb = KeyDBFactory(KeyDBWindow("snaps", self.keydb), encode_snapshot, partial(decode_snapshot, self.reverser))
    self.remotedb = KeyDBFactory(KeyDBWindow("remotes", self.keydb), encode_volume, decode_volume)
    self.check_userdata_blob = compose(invert, partial(_validate_checksum, self.reverser))

    exclude_file = Path('.farmignore', self.root)
    ignored = [safetype(self.mdd)]
    try:
        with exclude_file.open('rb') as exclude_fd:
          for raw_pattern in exclude_fd.readlines():
            pattern = ingest(raw_pattern.strip())
            excluded = safetype(Path(pattern, root))
            ignored.append(excluded)
    except IOError as e:
      if e.errno == NoSuchFile:
          pass
      else: raise e
    self.is_ignored = partial(skip_ignored, ignored)

  def csum_to_name(self, csum):
    """Return string name of link relative to udd"""
    #TODO someday when csums are parameterized, we inject the has params here.
    return _checksum_to_path(csum)

  def csum_to_path(self, csum):
    """Return absolute Path to a blob given a csum"""
    return Path(self.csum_to_name(csum), self.udd)

  def thawed(self, path):
    """Yield set of files not backed by FarmFS under path"""
    get_path = fmap(first)
    select_userdata_files = pipeline(
        ftype_selector([FILE]),
        get_path)
    return select_userdata_files(path.entries(self.is_ignored))

  def frozen(self, path):
    """Yield set of files backed by FarmFS under path"""
    get_path = fmap(first)
    select_userdata_files = pipeline(
        ftype_selector([LINK]),
        get_path)
    return select_userdata_files(path.entries(self.is_ignored))

  #NOTE: This assumes a posix storage engine.
  def freeze(self, path):
    assert isinstance(path, Path)
    assert isinstance(self.udd, Path)
    csum = path.checksum()
    blob = self.csum_to_path(csum)
    assert blob == self.udd.join(_checksum_to_path(csum))
    duplicate = blob.exists()
    if not duplicate:
      ensure_link(blob, path)
      ensure_readonly(blob)
    ensure_symlink(path, blob)
    ensure_readonly(path)
    return {"path":path, "csum":csum, "was_dup":duplicate}

  #Note: This assumes a posix storage engine.
  def thaw(self, user_path):
    assert isinstance(user_path, Path)
    csum_path = user_path.readlink()
    user_path.unlink()
    csum_path.copy(user_path)
    return user_path

  def repair_link(self, path):
    """Find all broken links and point them back at UDD"""
    assert(path.islink())
    oldlink = path.readlink()
    if oldlink.isfile():
        return
    csum = self.reverser(oldlink)
    newlink = self.csum_to_path(csum)
    assert newlink == Path(_checksum_to_path(csum), self.udd)
    if not newlink.isfile():
      raise ValueError("%s is missing, cannot relink" % newlink)
    else:
      path.unlink()
      path.symlink(newlink)
      return newlink

  def userdata_files(self):
    get_path = fmap(first)
    select_userdata_files = pipeline(
        ftype_selector([FILE]),
        get_path)
    return select_userdata_files(self.udd.entries())

  def link_checker(self):
    """Return a pipeline which given a list of SnapshotItems, returns the SnapshotItems with broken links to the blobstore"""
    select_links = partial(ifilter, lambda x: x.is_link())
    is_broken = lambda x: not self.blob_checker(x.csum())
    select_broken = partial(ifilter, is_broken)
    return pipeline(
            select_links,
            select_broken)

  def blob_checker(self, csum):
    """Returns true if the csum is in the store, false otherwise"""
    return self.csum_to_path(csum).exists()

  def trees(self):
    """Returns an iterator which contains all trees for the volume.
    The Local tree and all the snapshots"""
    tree = self.tree()
    snaps = imap(lambda x: self.snapdb.read(x), self.snapdb.list())
    return chain([tree], snaps)

  def items(self):
    """Returns an iterator which lists all SnapshotItems from all local snaps + the working tree"""
    return pipeline(
      concat)(self.trees())

  """Get a snap object which represents the tree of the volume."""
  def tree(self):
    tree_snap = TreeSnapshot(self.root, self.udd, self.is_ignored, reverser=self.reverser)
    return tree_snap

  """ Yield all the relative paths (safetype) for all the files in the userdata store."""
  def userdata_csums(self):
   # We populate counts with all hash paths from the userdata directory.
   for (path, type_) in self.udd.entries():
     assert isinstance(path, Path)
     if type_ == FILE:
       yield self.reverser(path)
     elif type_ == DIR:
       pass
     else:
       raise ValueError("%s is f invalid type %s" % (path, type_))

  def unused_blobs(self, items):
    """Returns the set of blobs not referenced in items"""
    select_links = partial(ifilter, lambda x: x.is_link())
    get_csums = fmap(lambda item: item.csum())
    referenced_hashes = pipeline(
            select_links,
            get_csums,
            uniq,
            set
            )(items)
    udd_hashes = set(self.userdata_csums())
    missing_data = referenced_hashes - udd_hashes
    assert len(missing_data) == 0, "Missing %s\nReferenced %s\nExisting %s\n" % (missing_data, referenced_hashes, udd_hashes)
    orphaned_csums = udd_hashes - referenced_hashes
    return orphaned_csums

  def delete_blob(self, csum):
      """Takes a csum, and removes it from the blobstore"""
      #TODO move to blobstore impl.
      blob_path = self.csum_to_path(csum)
      blob_path.unlink(clean=self.udd)

  """Yields similarity data for directories"""
  def similarity(self):
    tree = self.tree()
    dir_sigs = directory_signatures(tree, self.root)
    combos = combinations(dir_sigs.items(), 2)
    for ((dir_a, sigs_a), (dir_b, sigs_b)) in combos:
      intersection = len(sigs_a.intersection(sigs_b))
      count_a = len(sigs_a)
      count_b = len(sigs_b)
      yield (dir_a, count_a, dir_b, count_b, intersection)

@typed(FarmFSVolume, FarmFSVolume)
def tree_patcher(local_vol, remote_vol):
    return fmap(partial(tree_patch, local_vol, remote_vol))

def noop():
    pass

def blob_import(src_blob, dst_blob):
  """Takes source and destination as Paths"""
  # XXX Can't take csums because we don't know the segment format of remotes.
  if dst_blob.exists():
    return "Apply No need to copy blob, already exists"
  else:
    ensure_copy(dst_blob, src_blob)
    return "Apply Blob missing from local, copying"

@typed(FarmFSVolume, FarmFSVolume, SnapDelta)
def tree_patch(local_vol, remote_vol, delta):
  path = delta.path(local_vol.root)
  assert local_vol.root in path.parents(), "Tried to apply op to %s when root is %s" % (path, local_vol.root)
  if delta.csum is not None:
    dst_blob = local_vol.csum_to_path(delta.csum)
    src_blob = remote_vol.csum_to_path(delta.csum)
  else:
    dst_blob = None
    src_blob = None
  if delta.mode == delta.REMOVED:
    return (noop, partial(ensure_absent, path), ("Apply Removing %s", path))
  elif delta.mode == delta.DIR:
    return (noop, partial(ensure_dir, path), ("Apply mkdir %s", path))
  elif delta.mode == delta.LINK:
    blob_op = partial(blob_import, src_blob, dst_blob)
    tree_op = partial(ensure_symlink, path, dst_blob)
    tree_desc = ("Apply mklink %s -> " + delta.csum, path)
    return (blob_op, tree_op, tree_desc)
  else:
    raise ValueError("Unknown mode in SnapDelta: %s" % delta.mode)

#TODO yields lots of SnapDelta. Maybe in wrong file?
@typed(Snapshot, Snapshot)
def tree_diff(tree, snap):
  tree_parts = iter(tree)
  snap_parts = iter(snap)
  t = None
  s = None
  while True:
    if t == None:
      try:
        t = next(tree_parts)
      except StopIteration:
        pass
    if s == None:
      try:
        s = next(snap_parts)
      except StopIteration:
        pass
    if t is None and s is None:
      return # We are done!
    elif t is not None and s is not None:
      # We have components from both sides!
      if t < s:
        # The tree component is not present in the snap. Delete it.
        yield SnapDelta(t.pathStr(), SnapDelta.REMOVED)
        t = None
      elif s < t:
        # The snap component is not part of the tree. Create it
        yield SnapDelta(*s.get_tuple())
        s = None
      elif t == s:
        if t.is_dir() and s.is_dir():
          pass
        elif t.is_link() and s.is_link():
          if t.csum() == s.csum():
            pass
          else:
            change = t.get_dict()
            change['csum'] = s.csum()
            yield SnapDelta(t._path, t._type, s._csum)
        elif t.is_link() and s.is_dir():
          yield SnapDelta(t.pathStr(), SnapDelta.REMOVED)
          yield SnapDelta(s.pathStr(), SnapDelta.DIR)
        elif t.is_dir() and s.is_link():
          yield SnapDelta(t.pathStr(), SnapDelta.REMOVED)
          yield SnapDelta(s.pathStr(), SnapDelta.LINK, s.csum())
        else:
          raise ValueError("Unable to process tree/snap: unexpected types:", s.get_dict()['type'], t.get_dict()['type'])
        s = None
        t = None
      else:
        raise ValueError("Found pair that doesn't respond to > < == cases")
    elif t is not None:
      yield SnapDelta(t.pathStr(), SnapDelta.REMOVED)
      t = None
    elif s is not None:
      yield SnapDelta(*s.get_tuple())
      s = None
    else:
      raise ValueError("Encountered case where s t were both not none, but neither of them were none.")




from os import mkdir
from os import listdir
from os import link
from os import unlink
from os import symlink
from os import readlink
from os import rmdir
from os import stat
from os import chmod
from errno import ENOENT as FileDoesNotExist
from errno import EEXIST as FileExists
from errno import EISDIR as DirectoryExists
from hashlib import md5
from os.path import stat as statc
from os.path import normpath
from os.path import split
from os.path import isabs
from os.path import exists
from os.path import isdir
from shutil import copyfileobj
from os.path import isfile, islink, sep
from func_prototypes import typed, returned
from glob import fnmatch
from fnmatch import fnmatchcase
from functools import total_ordering, partial
from farmfs.util import ingest, safetype, uncurry, first
from future.utils import python_2_unicode_compatible
from safeoutput import open as safeopen
from filetype import guess, Type
import filetype
try:
    from itertools import ifilter
except ImportError:
    # On python3, filter is lazy.
    ifilter = filter

class XSym(Type):
    '''Implements OSX XSym link file type detector'''
    def __init__(self):
        super(XSym, self).__init__(
                mime='inode/symlink',
                extension='xsym')
    def match(self, buf):
        """Detects the MS-Dos symbolic link format from OSX.
        Format of XSym files taken from section 11.7.3 of Mac OSX Internals"""
        return (len(buf) >= 10 and
                buf[0] == 0x58 and # X
                buf[1] == 0x53 and # S
                buf[2] == 0x79 and # y
                buf[3] == 0x6d and # m
                buf[4] == 0xa and # \n
                buf[5] >= 0x30 and buf[5] <= 0x39 and # 0-9
                buf[6] >= 0x30 and buf[6] <= 0x39 and # 0-9
                buf[7] >= 0x30 and buf[7] <= 0x39 and # 0-9
                buf[8] >= 0x30 and buf[8] <= 0x39 and # 0-9
                buf[9] == 0xa # \n
                )
# XXX Dirty, we are touching the set of types in filetype package.
filetype.types.append(XSym())

_BLOCKSIZE = 65536

LINK=u'link'
FILE=u'file'
DIR=u'dir'

TYPES=[LINK, FILE, DIR]

#TODO should take 1 arg, return fn.
def skip_ignored(ignored, path, ftype):
  for i in ignored:
    if fnmatchcase(path._path, i):
      return True
  return False

def ftype_selector(keep_types):
  keep = lambda p, ft: ft in keep_types # Take p and ft since we may want to use it in entries.
  entry_keep = uncurry(keep) # Expand tuple from entries.
  entry_filter = partial(ifilter, entry_keep)
  return entry_filter

@total_ordering
@python_2_unicode_compatible
class Path:
  def __init__(self, path, frame=None):
    if path is None:
      raise ValueError("path must be defined")
    elif isinstance(path, Path):
      assert frame is None
      self._path = path._path
      self._parent = path._parent
    else:
      path = ingest(path)
      if frame is None:
        assert isabs(path), "Frame is required when building relative paths: %s" % path
        self._path = normpath(path)
      else:
        assert isinstance(frame, Path)
        assert not isabs(path), "path %s is required to be relative when a frame %s is provided" % (path, frame)
        self._path = frame.join(path)._path
      self._parent = first(split(self._path))
    assert isinstance(self._path, safetype)
    assert isinstance(self._parent, safetype)

  def __str__(self):
    return self._path

  def __repr__(self):
    return str(self)

  def mkdir(self):
    try:
      mkdir(self._path)
    except OSError as e:
      if e.errno == FileExists:
        pass
      elif e.errno == DirectoryExists:
        pass
      else:
        raise e

  # Returns the parent of self. If self is root ('/'), parent returns None.
  # You much check the output of parent before using the value.
  # Notcie that parent of root in the shell is '/', so this is a semantic difference
  # between us and POSIX.
  def parent(self):
    if self._path == sep:
      return None
    else:
      return Path(self._parent)

  def parents(self):
    paths = [self]
    path = self
    parent = path.parent()
    while parent is not None:
      paths.append(parent)
      path = parent
      parent = path.parent()
    return reversed(paths)

  #TODO This function returns leading '/' on relations.
  #TODO This function returns '/' for matches. It should return '.'
  #TODO This function doesn't handle "complex" relationships.
  #XXX This function leads to confusion. It returns a string when mostly you
  # want to mess with Paths. It should only be called in user output schenatios.
  #TODO Check where this is called and try to stop calling it.
  #TODO Rename this to somthing which disourages use.
  #TODO Rename this so the string return value is called out.
  def relative_to(self, relative, leading_sep=True):
    assert isinstance(relative, Path)
    if leading_sep == True:
      prefix = sep
    else:
      prefix = ""
    self_parents = list(self.parents())
    if relative in self_parents:
      relative_str = relative._path
      return prefix + self._path[len(relative_str)+1:]
    relative_parents = list(reversed(list(relative.parents())))
    if self in relative_parents:
      backups = relative_parents.index(self) - 1
      assert backups >= 0
      assert leading_sep == False, "Leading seperator is meaningless with backtracking"
      return sep.join([".."]*backups)
    raise ValueError("Relationship between %s and %s is complex" % (self, relative))


  def exists(self):
    return exists(self._path)

  def readlink(self, frame=None):
    return Path(readlink(self._path), frame)

  def link(self, dst):
    assert isinstance(dst, Path)
    link(dst._path, self._path)

  def symlink(self, dst):
    assert isinstance(dst, Path)
    symlink(dst._path, self._path)

  #TODO this behavior is the opposite of what one would expect.
  def copy(self, dst):
    assert isinstance(dst, Path)
    with open(self._path, 'rb') as src_fd:
      with safeopen(dst._path, 'wb') as dst_fd:
        copyfileobj(src_fd, dst_fd)

  def unlink(self, clean=None):
    try:
      unlink(self._path)
    except OSError as e:
      if e.errno == FileDoesNotExist:
        pass
    if clean is not None:
      parent = self.parent()
      parent._cleanup(clean)

  def rmdir(self, clean=None):
    rmdir(self._path)
    if clean is not None:
      parent = self.parent()
      parent._cleanup(clean)

  """Called on the parent of file or directory after a removal
  (if cleanup as asked for). Recuses cleanup until it reaches terminus.
  """
  def _cleanup(self, terminus):
    assert isinstance(terminus, Path)
    assert terminus in self.parents()
    if self == terminus:
      return
    if len(list(self.dir_gen())) == 0:
      self.rmdir(terminus)

  def islink(self):
    return islink(self._path)

  def isdir(self):
    return isdir(self._path)

  def isfile(self):
    return isfile(self._path)

  def checksum(self):
    hasher = md5()
    with self.open('rb') as fd:
      buf = fd.read(_BLOCKSIZE)
      while len(buf) > 0:
        hasher.update(buf)
        buf = fd.read(_BLOCKSIZE)
      digest = safetype(hasher.hexdigest())
      return digest

  def __cmp__(self, other):
    return (self > other) - (self < other)

  def __eq__(self, other):
    assert isinstance(other, Path)
    return self._path == other._path

  def __ne__(self, other):
    assert isinstance(other, Path)
    return not (self == other)

  def __lt__(self, other):
    assert isinstance(other, Path)
    return self._path.split(sep) < other._path.split(sep)

  def __hash__(self):
    return hash(self._path)

  def join(self, child):
    child = safetype(child)
    try:
      output = Path( self._path + sep + child)
    except UnicodeDecodeError as e:
      raise ValueError(str(e) + "\nself path: "+ self._path + "\nchild: ", child)
    return output

  def dir_gen(self):
    """Generates the set of Paths under this directory"""
    assert self.isdir(), "%s is not a directory" % self._path
    assert isinstance(self._path, safetype)
    names = listdir(self._path)
    for name in names:
      child = self.join(name)
      yield child

  def ftype(self):
    if self.islink():
      return LINK
    elif self.isfile():
      return FILE
    elif self.isdir():
      return DIR
    else:
      raise ValueError("%s is not in %s" % (self, types))

  def entries(self, skip=None):
    t = self.ftype()
    if skip and skip(self, t):
      return
    yield (self, t)
    if t == DIR:
      children = self.dir_gen()
      for dir_entry in sorted(children):
        for x in dir_entry.entries(skip):
          yield x

  def open(self, mode):
    return open(self._path, mode)

  def stat(self):
    return stat(self._path)

  def chmod(self, mode):
    return chmod(self._path, mode)

  def filetype(self):
    # XXX Working around bug in filetype guess.
    # Duck typing checks don't work on py27, because of str bytes confusion.
    # So we read the file outselves and put it in a bytearray.
    # Remove this when we drop support for py27.
    with self.open("rb") as fd:
      type = guess(bytearray(fd.read(256)))
      if type:
          return type.mime
      else:
          return None

@returned(Path)
def userPath2Path(arg, frame):
    """
    Building paths using conventional POSIX systems will discard CWD if the
    path is absolute. FarmFS makes passing of CWD explicit so that path APIs
    are pure functions. Additionally FarmFS path construction doesn't allow
    for absolute paths to be mixed with frames. This is useful for
    spotting bugs and making sure that pathing has strong guarantees. However
    this comes at the expense of user expectation. When dealing with user
    input, there is an expecation that POSIX semantics are at play.
    userPath2Path checks to see if the provided path is absolute, and if not,
    adds the CWD frame.
    """
    arg = ingest(arg)
    if isabs(arg):
      return Path(arg)
    else:
      return Path(arg, frame)

@returned(bool)
@typed(Path)
def target_exists(link):
  assert link.islink()
  target = link.readlink(link.parent())
  return target.exists()

#TODO this function is dangerous. Would be better if we did sorting in the snaps to ensure order of ops explicitly.
@typed(Path)
def ensure_absent(path):
  if path.exists():
    if path.isdir():
      for child in path.dir_gen():
        ensure_absent(child)
      path.rmdir()
    else:
      path.unlink()
  else:
    pass # No work to do.

@typed(Path)
def ensure_dir(path):
  if path.exists():
    if path.isdir():
      pass # There is nothing to do.
    else:
      path.unlink()
      path.mkdir()
  else:
    assert path != ROOT, "Path is root, which must be a directory"
    parent = path.parent()
    assert parent != path, "Path and parent were the same!"
    ensure_dir(parent)
    path.mkdir()

@typed(Path, Path)
def ensure_link(path, orig):
  assert orig.exists()
  parent = path.parent()
  assert parent != path, "Path and parent were the same!"
  ensure_dir(parent)
  ensure_absent(path)
  path.link(orig)

write_mask = statc.S_IWUSR | statc.S_IWGRP | statc.S_IWOTH
read_only_mask = ~write_mask

@typed(Path)
def ensure_readonly(path):
  mode = path.stat().st_mode
  read_only = mode & read_only_mask
  path.chmod(read_only)

@typed(Path)
def is_readonly(path):
  mode = path.stat().st_mode
  writable = mode & write_mask
  return bool(writable)

@typed(Path, Path)
def ensure_copy(dst, src):
  assert src.exists()
  parent = dst.parent()
  assert parent != dst, "dst and parent were the same!"
  ensure_dir(parent)
  ensure_absent(dst)
  src.copy(dst)

@typed(Path, Path)
def ensure_symlink(path, orig):
  assert orig.exists()
  ensure_symlink_unsafe(path, orig._path)

@typed(Path, safetype)
def ensure_symlink_unsafe(path, orig):
  parent = path.parent()
  assert parent != path, "Path and parent were the same!"
  ensure_dir(parent)
  ensure_absent(path)
  symlink(orig, path._path)

"""
Creates/Deletes directories. Does whatever is required inorder
to make and open a file with the mode previded.

Mode settings to consider are:
 O_CREAT         create file if it does not exist
 O_TRUNC         truncate size to 0
 O_EXCL          error if O_CREAT and the file exists
"""
def ensure_file(path, mode):
  assert isinstance(path, Path)
  parent = path.parent()
  assert parent != path, "Path and parent were the same!"
  ensure_dir(parent)
  fd = path.open(mode)
  return fd

ROOT = Path(sep)




# Worked out from https://raganwald.com/2017/04/30/transducers.html
# JS -> Python conversion.

def reduceWith(reducer, seed, iterable):
    """
    reduceWith takes reducer as first argument, computes a reduction over iterable.
    Think foldl from Haskell.
    reducer is (b -> a -> b)
    Seed is b
    iterable is [a]
    reduceWith is (b -> a -> b) -> b -> [a] -> b
    """
    accumulation = seed
    for value in iterable:
        accumulation = reducer(accumulation, value)
    return accumulation

arrayOf = lambda acc, val: acc.append(val) or acc
arrayOf.__doc__ = \
"""
Optimized version of array accumulator which doesn't reallocate on every loop
iteration.
"""

sumOf = lambda acc, val: acc + val
sumOf.__doc__ = """Reducer which computes a sum"""

def joinedWith(seperator):
    def joint(acc, val):
        if acc == '':
            return val
        else:
            return "%s%s%s" % (acc, seperator, val)
    return joint

map = lambda fn: lambda reducer: lambda acc, val: reducer(acc, fn(val))
map.__doc__ = """map is decorator which parameterizes the (+1) as a parameter."""

filter = lambda pred: \
        lambda reducer: \
        lambda acc, val: reducer(acc, val) if pred(val) else acc
filter.__doc__ =  \
"""
pred is (a->Bool)
reducer is (b -> a -> b)
"""

"""
How can we perform an arbitrary series of compositions?
Yes, with a reduction!
"""
compositionOf = lambda acc, val: lambda *args, **kwargs: val(acc(*args, **kwargs))
compose = lambda *fns: reduceWith(compositionOf, lambda x: x, fns);

def transduce(transformer, reducer, seed, iterable):
    """
    transformer is (a -> b)
    reducer is (b -> a -> b)
    seed is b
    iterable is [a]
    """
    transformedReducer = transformer(reducer)
    accumulation = seed
    for value in iterable:
        accumulation = transformedReducer(accumulation, value)
    return accumulation




#!/usr/bin/env python
import sys
from setuptools import setup
from setuptools.command.test import test as TestCommand

requires = ['func_prototypes>=0.5.0', 'docopt', 'delnone', 'future', 'safeoutput>=2.0', 'filetype==1.0.6', "S3Lib==1.1.1"]

test_requires = ['tox', 'pytest==4.6.8', 'tabulate']

setup(
    name='farmfs',
    version='0.8.1',
    author='Andrew Thomson',
    author_email='athomsonguy@gmail.com',
    packages=['farmfs'],
    install_requires = requires,
    entry_points = {
      'console_scripts': [
        'farmfs = farmfs.ui:ui_main',
        'farmdbg = farmfs.ui:dbg_main',
        ],
    },
    url='http://github.com/andrewguy9/farmfs',
    license='MIT',
    description='tool which de-duplicates files in a filesystem by checksum.',
    long_description_content_type='text/markdown',
    long_description=open('README.md').read(),
    scripts=['bin/snap.sh'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Topic :: System :: Filesystems',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 2.7',
    ],
)



from farmfs.fs import Path, LINK, DIR, FILE, ingest
from func_prototypes import typed
from delnone import delnone
from os.path import sep
from functools import total_ordering
from farmfs.util import safetype
from future.utils import python_2_unicode_compatible

try:
    from itertools import imap
except ImportError:
    # On python3 map is lazy.
    imap = map

@total_ordering
@python_2_unicode_compatible
class SnapshotItem:
  def __init__(self, path, type, csum=None):
    assert isinstance(type, safetype)
    assert type in [LINK, DIR], type
    if (isinstance(path, Path)):
        path = path._path #TODO reaching into path.
    assert isinstance(path, safetype)
    if type == LINK:
      if csum is None:
        raise ValueError("checksum should be specified for links")
    self._path = ingest(path)
    self._type = ingest(type)
    self._csum = csum and ingest(csum) # csum can be None.

  #TODO create a path comparator. cmp has different semantics.
  def __cmp__(self, other):
    assert other is None or isinstance(other, SnapshotItem)
    if other is None:
      return -1
    self_path = Path(self._path)
    other_path = Path(other._path)
    return self_path.__cmp__(other_path)

  def __eq__(self, other):
    return self.__cmp__(other) == 0

  def __ne__(self, other):
    return self.__cmp__(other) != 0

  def __lt__(self, other):
    return self.__cmp__(other) < 0

  def get_tuple(self):
    return (self._path, self._type, self._csum)

  def get_dict(self):
    return delnone(dict(path=self._path,
            type=self._type,
            csum=self._csum))

  def pathStr(self):
    assert isinstance(self._path, safetype)
    return self._path;

  def is_dir(self):
    return self._type == DIR

  def is_link(self):
    return self._type == LINK

  def csum(self):
    assert self._type == LINK, "Encountered unexpected type %s in SnapshotItem for path %s" % (self._type, self._path)
    return self._csum

  def __str__(self):
    return "<%s %s %s>" % (self._type, self._path, self._csum)

  def to_path(self, root):
    return root.join(self._path)

class Snapshot:
  pass

class TreeSnapshot(Snapshot):
  def __init__(self, root, udd, is_ignored, reverser):
    assert isinstance(root, Path)
    self.root = root
    self.udd = udd
    self.is_ignored = is_ignored
    self.reverser = reverser
    self.name = '<tree>'

  def __iter__(self):
    root = self.root
    udd = self.udd
    def tree_snap_iterator():
      last_path = None # Note: last_path is just used to debug snapshot order issues. Remove once we have confidence.
      for path, type_ in root.entries(self.is_ignored):
        if last_path:
          assert last_path < path, "Order error: %s < %s" % (last_path, Path)
        last_path = path
        if type_ == LINK:
          ud_str = self.reverser(path.readlink().relative_to(udd))
        elif type_ == DIR:
          ud_str = None
        elif type_ == FILE:
          continue
        else:
          raise ValueError("Encounted unexpected type %s for path %s" % (type_, entry))
        yield SnapshotItem(path.relative_to(root), type_, ud_str)
    return tree_snap_iterator()

class KeySnapshot(Snapshot):
  def __init__(self, data, name, reverser):
    assert(data)
    self.data = data
    self._reverser = reverser
    self.name = name

  def __iter__(self):
    def key_snap_iterator():
      assert(self.data)
      for item in self.data:
        if isinstance(item, list):
          assert len(item) == 3
          (path_str, type_, ref) = item
          assert isinstance(path_str, safetype)
          assert isinstance(type_, safetype)
          if ref is not None:
            csum = self._reverser(ref)
            assert isinstance(csum, safetype)
          else:
            csum = None
          parsed = SnapshotItem(path_str, type_, csum)
        elif isinstance(item, dict):
          parsed = SnapshotItem(**item)
        yield parsed
    return iter(sorted(key_snap_iterator()))

@python_2_unicode_compatible
class SnapDelta:
  REMOVED=u'removed'
  DIR=DIR
  LINK=LINK
  _modes = [REMOVED, DIR, LINK]
  def __init__(self, pathStr, mode, csum=None):
    assert isinstance(pathStr, safetype), "didn't expect type %s" % type(pathStr)
    assert isinstance(mode, safetype) and mode in self._modes
    if mode == self.LINK:
      # Make sure that we are looking at a csum, not a path.
      assert csum is not None and csum.count(sep) == 0
    else:
      assert csum is None
    self._pathStr = pathStr
    self.mode = mode
    self.csum = csum

  def path(self, root):
    return root.join(self._pathStr)

  def __str__(self):
      return "{"+self.path("")+","+self.mode+","+self.csum+"}" # Not a great encoding.

def encode_snapshot(snap):
  return list(imap(lambda x: x.get_dict(), snap))

def decode_snapshot(splitter, reverser, data, key):
  return KeySnapshot(data, key, splitter, reverser)

